/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include "monitor.h"

#define REC_LEN     1024

using namespace std;
/**
 * Return file path
 */
string vhost::filepath()
{
   return this->fields[vhost::FILE_FIELD];
}
/**
 * Return file path in a form suitable for open
 */
const char* vhost::filestr()
{
   return this->filepath().c_str();
}
/**
 * Parse a log record. The 0th slot of the rec array is used as the parse space for
 * incremental parsing. Each new record is counted as a 'hit'
 */
void vhost::parseLogRecord(topvhost *app,char *s, int ct)
{
   char              ch;
   string            key;
   int               n;

   this->meter += ct;
   while(ct--) {
    switch(ch=*s++) {
      case '"':
         this->iq = !this->iq;
         break;
      case ' ':
        if (this->iq || this->fields[0][0]=='[') {
            this->fields[0] += ch;
            break;
            }
         else if (!this->fields[0].length())
            break;
      case '\n':
   done:
         n = app->xfer[this->argn];
         if (n>0)
            this->fields[n] = this->fields[0];
         if (ch=='\n') {
            this->argn = 0;
            this->hits++;
            }
         else this->argn++;
         this->fields[0] = "";
         break;
      case ']':
         this->fields[0] += ch;
         if (!this->iq && this->fields[0][0]=='[')
            goto done;
         break;
      default:
         this->fields[0] += ch;
      }
    }
}
/**
 * Clean the dirty bit
 */
bool vhost::setClean(topvhost *app)
{
   char  buf[REC_LEN];
   int   ct;

   int fh = open(this->filestr(),O_RDONLY);
   if (fh!=-1) {
       if (lseek(fh,this->fsz,SEEK_SET)!=-1) {
         while((ct = read(fh,buf,sizeof(buf)))>0)
           this->parseLogRecord(app,buf,ct);
         this->fsz = lseek(fh,0,SEEK_CUR);
         }
      close(fh);
      return true;
      }
   return false;
}
/**
 * Save info for later processing
 */
bool vhost::setDirty(bool init)
{
   struct stat attrib;

   if (stat(this->filestr(),&attrib)==0) {
      this->ldtm  = attrib.st_mtime;
      this->dirty = init;
      if (!init)
         this->fsz = attrib.st_size;
      return true;
      }
   return false;
}
/**
 * Format vhost content using a sprintf like specification using the place holders:
 *
 *  %H  = hour
 *  %M  = minute
 *  %S  = second
 *  %Y  = year
 *  %d  = day
 *  %h  = remote host     - right aligned and truncated/padded to width
 *  %m  = month
 *  %o  = count (hits)
 *  %r  = request string  - left aligned and truncated/padded to width
 *  %v  = domain name     - left aligned and always padded to namelength
 */
void vhost::setString(topvhost *app)
{
   char buf[128];
   struct tm *ts = localtime(&this->ldtm);
   string token,val;
   string::iterator fmt = app->rfmt.begin();
   int ch,idx,iv;
   unsigned i, width;

   app->tline = "";
   idx = vhost::OUTPUT_FIELDS;
   while( fmt != app->rfmt.end() )
      if ((ch = *fmt++)=='%') {
         token = ch;
         if (strchr("-+#0 ",*fmt)!=NULL)           // flags
            token += *fmt++;
         while(strchr("0123456789.",*fmt)!=NULL)   // width - precision
            token += *fmt++;
         if (*fmt == '{') {
            do    *fmt++;
            while (fmt != app->rfmt.end() && ch != '}');
            }
         switch(ch = *fmt++) {
            case 'H':   iv = ts->tm_hour;    break;
            case 'M':   iv = ts->tm_min;     break;
            case 'S':   iv = ts->tm_sec;     break;
            case 'Y':   iv = ts->tm_year;    break;
            case 'd':   iv = ts->tm_mday;    break;
            case 'm':   iv = ts->tm_mon+1;   break;
            case 'o':   iv = this->hits;     break;
            default:    // Cehirsv
               val   = this->fields[idx++];
               token.erase(0,1);
               width = ch=='v'? app->namelen:atoi(token.c_str());
               for(i=0;i<val.length();i++)
                  if (i>= width)
                     val = val.substr(0,i);
                  else switch(val[i]) {
                     case '\b':  case '\t':
                        val[i] = ' ';
                        break;
                     case '\r':  case '\n':
                        val = val.substr(0,i);
                        break;
                     }
               app->tline += val;
               while(i++<width) app->tline += ' ';
               continue;
            }
         token += 'd';
         snprintf(buf,sizeof(buf),token.c_str(), iv);
         app->tline += buf;
         }
      else app->tline += ch;
}
/**
 * Process row format to produce and array of needed input tokens. This is needed
 * to assign slots for the values in the host field vector.
 */
void vhost::setVector(topvhost *app)
{
   string key;
   string::iterator fmt = app->rfmt.begin();
   int ch;

   while( fmt != app->rfmt.end() )
      if ((ch = *fmt++)=='%') {
         if (strchr("-+#0 ",*fmt)!=NULL)           // flags
            fmt++;
         while(strchr("0123456789.",*fmt)!=NULL)   // width - precision
            fmt++;
         key = "";
         if (*fmt == '{') {
            do    key += ch = *fmt++;
            while (fmt != app->rfmt.end() && ch != '}');
            }
         switch(ch = *fmt++) {
            case 'C':   case 'e':   case 'h':   case 'i':   case 'r':   case 's':
            case 'v':
               key += ch;
               app->needs.push_back(key);
            }
         }
}
